#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "d3des.h"
#include "VncClient.h"

#ifdef LOCAL_BE
# define BE_LOCAL_32(a) (a)
# define BE_LOCAL_16(a) (a)
#else
# define BE_LOCAL_32(a) ((((a)&0xff)<<24)|(((a)>>24)&0xff)|(((a)&0xff00)<<8)|(((a)>>8)&0xff00))
# define BE_LOCAL_16(a) ((((a)&0xff)<<8)|(((a)>>8)&0xff))
#endif

VncClient::VncClient(FbHandler fbHandler, ErrorHandler errorHandler) :  handleFb(fbHandler), handleError(errorHandler)
{
	
	rbufSize = VNCCLIENT_RBUFIZE;
	sbufSize = VNCCLIENT_SBUFSIZE;
	
	allocBuffers();
}

VncClient::~VncClient()
{
	close();
}

void VncClient::allocBuffers()
{
	rbuf = (char*)malloc(rbufSize);
	sbuf = (char*)malloc(sbufSize);
	if(!rbuf || !sbuf) { throwError(ERROR_FATAL, "Unable to allocate buffer, not enough memory."); }
}

void VncClient::connect(const char *server, unsigned long port)
{
	struct hostent *shost;
	
	shost = gethostbyname(server);
	if(!shost) { throwError(ERROR_FATAL, "Host unreachable...");  return; }

	struct sockaddr_in saddr;
	
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	memcpy(&saddr.sin_addr, shost->h_addr, sizeof(unsigned long));
	
	fd = socket(AF_INET, SOCK_STREAM, 0);
	
	int i;
	
	i = ::connect(fd, (const struct sockaddr*)&saddr, sizeof(struct sockaddr_in));
	if(i != 0) { throwError(ERROR_FATAL, "Unable to connect host..."); return; }	
}

void VncClient::close()
{
	::close(fd);
}

void VncClient::handshake()
{
	int i;
	int versionMajor, versionMinor;
	
	forceReceive();
	i = sscanf(rbuf, rfbProtocolVersionFormat, &versionMajor, &versionMinor);
	if(i != 2) { throwError(ERROR_WARNING, "Bad version message from server."); return; }
	
	if(versionMajor != 3)  { throwError(ERROR_WARNING, "Server is not RFB 003."); return; }
	
	rfbVersion = versionMinor;
		
	log("Server version : " rfbProtocolVersionFormat, versionMajor, versionMinor);
	
	sprintf(sbuf, rfbProtocolVersionFormat, rfbProtocolMajorVersion, rfbVersion);
	send(sz_rfbProtocolVersionMsg);
	
	forceReceive();
	
	if(rfbVersion >= 7)
	{
		if(*(uint8_t*)rbuf == 0) { throwError(ERROR_FATAL, "Server version is incompatible."); return; }
		*(uint8_t*)sbuf = rfbSecTypeNone;
		send(1);
		
		if(rfbVersion >= 8)
		{
			forceReceive();
			if(*(uint32_t*)rbuf != 0) { throwError(ERROR_FATAL, "Authentication failed."); return; }
		}
	}
	else
	{
		switch(*(uint32_t*)rbuf)
		{
			case rfbSecTypeInvalid: throwError(ERROR_FATAL, "Server version is incompatible."); return;
				break;
			case rfbSecTypeNone:
				break;
			case rfbSecTypeVncAuth: throwError(ERROR_FATAL, "VNC Authentication asked."); return;
				break;
		}
	}
	
	((rfbClientInitMsg*)sbuf)->shared = 1;
	send(sz_rfbClientInitMsg);
	
	forceReceive();
	
	remoteWidth = BE_LOCAL_16( ((rfbServerInitMsg*)rbuf)->framebufferWidth );
	remoteHeight = BE_LOCAL_16( ((rfbServerInitMsg*)rbuf)->framebufferHeight );
	
	
	copyRfbString(remoteName, &(((rfbServerInitMsg*)rbuf)->nameLength));
	
	mode = MODE_NORMAL;
	
}

void VncClient::setPixelFormat(PixelFormat *format)
{
	((rfbSetPixelFormatMsg*)sbuf)->format.bitsPerPixel = format->depth;
	((rfbSetPixelFormatMsg*)sbuf)->format.depth = format->depth;
#ifdef LOCAL_BE
	((rfbSetPixelFormatMsg*)sbuf)->format.bigEndian = 1;
#else
	((rfbSetPixelFormatMsg*)sbuf)->format.bigEndian = 0;
#endif
	((rfbSetPixelFormatMsg*)sbuf)->format.trueColour = 1;
	((rfbSetPixelFormatMsg*)sbuf)->format.redMax = BE_LOCAL_16(format->redMax);
	((rfbSetPixelFormatMsg*)sbuf)->format.greenMax = BE_LOCAL_16(format->greenMax);
	((rfbSetPixelFormatMsg*)sbuf)->format.blueMax = BE_LOCAL_16(format->blueMax);
	((rfbSetPixelFormatMsg*)sbuf)->format.redShift = format->redShift;
	((rfbSetPixelFormatMsg*)sbuf)->format.blueShift = format->blueShift;
	((rfbSetPixelFormatMsg*)sbuf)->format.greenShift = format->greenShift;
			
	((rfbSetPixelFormatMsg*)sbuf)->type = rfbSetPixelFormat;

	
	send(sz_rfbSetPixelFormatMsg);
	
}

void VncClient::setEncodings()
{
	((rfbSetEncodingsMsg*)sbuf)->type = rfbSetEncodings;
	((rfbSetEncodingsMsg*)sbuf)->nEncodings = BE_LOCAL_16(1);
	*((uint32_t*)(sbuf + sz_rfbSetEncodingsMsg)) = BE_LOCAL_32(rfbEncodingRaw);
	
	send(sz_rfbSetEncodingsMsg + 1 * sizeof(uint32_t));
	
}

void VncClient::framebufferUpdateRequest(Rect *rect, bool updateOnly)
{
	
	((rfbFramebufferUpdateRequestMsg*)sbuf)->type = rfbFramebufferUpdateRequest;
	
	((rfbFramebufferUpdateRequestMsg*)sbuf)->incremental = updateOnly ? 1 : 0;
	
	((rfbFramebufferUpdateRequestMsg*)sbuf)->x = BE_LOCAL_16( rect->x );
	((rfbFramebufferUpdateRequestMsg*)sbuf)->y = BE_LOCAL_16( rect->y );
	((rfbFramebufferUpdateRequestMsg*)sbuf)->w = BE_LOCAL_16( rect->w );
	((rfbFramebufferUpdateRequestMsg*)sbuf)->h = BE_LOCAL_16( rect->h );
	
	send(sz_rfbFramebufferUpdateRequestMsg);
	
}

void VncClient::update()
{
	int r;

	r = pollReceive();
	
	if(r < 0) { return; }
	else if(r == 0) { throwError(ERROR_FATAL, "Connection reset by peer."); return; }
	
	switch(mode)
	{
		case MODE_NORMAL:
		{
			rfbServerToClientMsg *msg = (rfbServerToClientMsg*)rbuf;
			switch(msg->type)
			{
				case rfbFramebufferUpdate:
			
					fbRectangles = BE_LOCAL_16( msg->fu.nRects );
					mode = MODE_FB;
					fbDataCount = remoteWidth * remoteHeight * 3;
			
					break;
				case rfbSetColourMapEntries:
					break;
				case rfbBell:
					break;
				case rfbServerCutText:
					char tmp[512];
					copyRfbString(tmp, &msg->sct.length);
					throwError(ERROR_WARNING, "CutText from server: %s", tmp);
					break;	
				case rfbResizeFrameBuffer:
					break;
				case rfbKeyFrameUpdate:
					break;
				case rfbReSizeFrameBuffer:
					break;
				default:
					throwError(ERROR_WARNING, "Unknown message type (%X) from server...", *(uint8_t*)rbuf); return;
			}	
			break;
		}
		case MODE_FB:
		{
			rfbFramebufferUpdateRectHeader *rect = (rfbFramebufferUpdateRectHeader*)rbuf;
			
			log("Rect X: %d Y: %d W: %d H: %d Encoding: %d", rect->r.x, rect->r.y, rect->r.w, rect->r.h, rect->encoding);
			
			switch(BE_LOCAL_32( rect->encoding ))
			{
				case rfbEncodingRaw:
				{
					int32_t i = 0;
					while(i < (BE_LOCAL_32(rect->r.w) * BE_LOCAL_32(rect->r.h) * 4))
					{
// 						r = forceReceive();
// 						handleFb(this, rbuf, r);
						i += r;
					}
					break;
				}
				default:
					throwError(ERROR_WARNING, "Unsupported encoding %X", rect->encoding);
					break;
			}
			
			break;
		}
	}
	
	//throwError(ERROR_WARNING, "Incoming message type %X.", *(uint8_t*)rbuf); return;
}

int VncClient::receive()
{
	int r;
	r = ::recv(fd, rbuf, rbufSize, 0);
	
#ifdef DEBUG
	if(r > 0) {
		char tmp[512] = {0};
		int i;
		for(i = 0; i < (r > 128 ? 128 : r) ; i++) { sprintf(tmp, "%s %02X", tmp, rbuf[i] & 0xff); }
		sprintf(tmp, "%s | ", tmp);
		for(i = 0; i < (r > 128 ? 128 : r); i++) { char c = rbuf[i] & 0xff; sprintf(tmp, "%s%c", tmp, c >= ' ' ? c : '.'); }
		log("R%d %s\n", r, tmp);
	}
#endif
	
	return r;
}

int VncClient::forceReceive()
{
	int r;
	do { r = receive(); } while(r <= 0);
	return r;
}

int VncClient::pollReceive()
{
	int r;
	SocketSetBlocking(fd, true);
	r = receive();
	SocketSetBlocking(fd, false);
	return r;
}

int VncClient::send(unsigned int size)
{
	int s;
	s = ::send(fd, sbuf, size < sbufSize ? size : sbufSize, 0);
	
#ifdef DEBUG
	char tmp[512] = {0};
	int i;
	for(i = 0; i < s; i++) { sprintf(tmp, "%s %02X", tmp, sbuf[i] & 0xff); }
	sprintf(tmp, "%s | ", tmp);
	for(i = 0; i < s; i++) { char c = sbuf[i] & 0xff; sprintf(tmp, "%s%c", tmp, c >= ' ' ? c : '.'); }
	log("S%d %s\n", s, tmp);
#endif
	
	return s;
}

void VncClient::log(const char *msg, ...)
{
	va_list vl;
	va_start(vl, msg);
	vfprintf(stderr, msg, vl);
	va_end(vl);
}

void VncClient::throwError(ErrorType type, const char *msg, ...)
{
	va_list vl;
	va_start(vl, msg);
	vsprintf(currentError.msg, msg, vl);
	va_end(vl);
	currentError.type = type;
	if(handleError) handleError(this);
}

VncClient::ErrorType VncClient::catchError(char *buffer)
{
	ErrorType type = currentError.type;
	if(type != ERROR_NONE)
	{
		strcpy(buffer, currentError.msg);
		currentError.type = ERROR_NONE;
		return type;
	}
	return ERROR_NONE;
}

void VncClient::copyRfbString(char *dst, const void *src)
{
	unsigned int s = BE_LOCAL_32(*(uint32_t*)src);
	memcpy(dst, (char*)src + sizeof(uint32_t), s);
	dst[s] = '\0';
}
