/*
 *  Copyright (C) 2006 Thoduv.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ini.h"

ini_file_t *ini_blank_new()
{
	ini_file_t *ini;
	
	ini = (ini_file_t*)malloc(sizeof(ini_file_t));	/* allocate new struct */
	if(!ini) { return NULL; }
	ini->sec = NULL;	/* no sections */
	
	return ini;
}

ini_file_t *ini_parse(const char *filename)
{
	char line[512];
	char *p, *q, *r;
	FILE *file;
	ini_file_t *ini;
	ini_section_t *sec = NULL, *sec_stack = NULL;
	ini_value_t *val = NULL, *val_stack = NULL;
	
	ini = ini_blank_new();
	if(!ini) return NULL;
	
	file = fopen(filename, "r");	/* open file */
	if(!file) { return NULL; }
	
	do
	{
		fgets(line, 512, file);	/* read a line */
		p = line;
		
		while(*p == ' ') p++;	/* avoid spaces */
		if(*p == '\r' || *p == '\n') continue;		/* blank line */
		
		if(*p == '#' || *p == ';') continue;	/* comment */
		
		if(*p == '[')	/* new section */
		{
			p++;
			q = p;
			while(*q != '\0' && *q != '\r' && *q != '\n')
			{
				if(*q == ']')	/* section name end reached */
				{
					sec = (ini_section_t*)malloc(sizeof(ini_section_t));	/* allocate new section */
					if(!sec) { fclose(file); return NULL; }
					if(sec_stack) { sec_stack->next = sec; }	/* add section in chain (if chain exist) */
					else { ini->sec = sec; }	/* if chain does not exist, it's the first section (register it as the beginning of chain) */
					sec_stack = sec;	/* set top of chain to new section */
					
					strncpy(sec->name, p, q-p);	/* copy section name */
					sec->name[q-p] = '\0';
					sec->next = NULL;	/* this section is the last in the chain */
					sec->val = NULL;	/* no val for now */
					val_stack = NULL;		/* begin new stack for values */
					break;
				}
				q++;
			}
		}
		else if(sec_stack) /* so, it may be a value, but we accept value only if we are in a section */
		{
			q = strchr(p, '=');
			if(q) { r = q-1; while(*r == ' ' && (r-p) > 0) r--; }
			if(q && r-p > 0)	/* there is an equal somewhere, so it's a value */
			{
				val = (ini_value_t*)malloc(sizeof(ini_value_t));	/* allocate new value */
				if(!val) { fclose(file); return NULL; }
				if(val_stack) {  val_stack->next = val; }	/* add section in chain (if chain exist) */
				else { sec_stack->val = val; } /* if chain does not exist, it's the first value of the section (register it as the beginning of chain) */
				val_stack = val; /* set top of chain to new section */
				
				r = q-1;
				while(*r == ' ' && (r-p) > 0) r--;	/* avoid spaces before equal */
				strncpy(val->name, p, r-p+1);	/* copy value name */
				val->name[r-p+1] = '\0';
				r = q+1;
				while(*r == ' ' && *r != '\r' && *r != '\n' && *r != '\0') r++;	/* avoid spaces after equal */
				p = val->value;
				while(*r != '\r' && *r != '\n' && *r != '\0' && p-val->value < 256) { *p = *r; p++; r++; }
				val->next = NULL;	/* this section is the last in the chain */
			}
		}
		
	}
	while(!feof(file));	/* check if it's the end of the file */
	
	fclose(file);	/* we're done, close file */
	
	return ini;	
}

static void ini_free_section(ini_section_t *sec)
{
	ini_value_t *val, *old_val;
	val = sec->val;
	while(val)
	{
		old_val = val;
		val = val->next;
		free(old_val);	/* free this value */
	}
}

void ini_free_file(ini_file_t *ini)
{
	ini_section_t *sec, *old_sec;
	
	sec = ini->sec;
	while(sec)
	{
		ini_free_section(sec);	
		old_sec = sec;
		sec = sec->next;
		free(old_sec);	/* free this section */
	}
	free(ini);	/* free main struct */
}

/* case insensitive string comparison */
int strcmp_ncase(const char *s1, const char *s2)
{
#define _to_lower(x) (((x) >= 'A' && (x) <= 'Z') ? ((x) - 'A' + 'a') : (x))
#define _cmp_ncase(a, b) (_to_lower(a) == _to_lower(b))
	const char *p1 = s1;
	const char *p2 = s2;
	while(*p1 != '\0' || *p2 != '\0') {	if(!_cmp_ncase(*p1, *p2)) { return -1; } p1++; p2++; }
	if(*p1 == '\0' && *p2 == '\0') return 0;
	return -1;
#undef _to_lower
#undef _cmp_ncase
}

ini_section_t *ini_req_section(ini_file_t *ini, const char *secs)
{
	ini_section_t *sec;
	
	sec = ini->sec;
	while(sec)
	{
		if(strcmp_ncase(secs, sec->name) == 0)	/* if it's the section we're searching */
		{
			return sec;
		}
		sec = sec->next;
	}
	return NULL;
}

const char *ini_req_value(ini_file_t *ini, const char *secs, const char *vals)
{
	ini_section_t *sec;
	ini_value_t *val;
	
	sec = ini->sec;
	while(sec)
	{
		if(strcmp_ncase(secs, sec->name) == 0)	/* if it's the section we're searching */
		{
			val = sec->val;
			while(val)
			{
				if(strcmp_ncase(vals, val->name) == 0)	/* if it's the value we're searching */
				{
					return val->value;
				}
				val = val->next;
			}
		}
		sec = sec->next;
	}
	return NULL;
}

ini_section_t *ini_end_section(ini_section_t *sec)	/* returns end of a section chain */
{
	ini_section_t *sec_stack = sec, *sec_stack_old = NULL;
	while(sec_stack)
	{
		sec_stack_old = sec_stack;
		sec_stack = sec_stack->next;
	}
	return sec_stack_old;
}

void ini_add_section(ini_file_t *ini, const char *secs)
{
	ini_section_t *sec, *sec_end = NULL;
	
	sec = (ini_section_t*)malloc(sizeof(ini_section_t));	/* alloc new section */
	if(!sec) return;
	strncpy(sec->name, secs, 64);
	sec->next = NULL;
	sec->val = NULL;
	
	if(ini->sec)	/*check  if chain exists */
	{
		sec_end = ini_end_section(ini->sec);	/* get end of chain */
		sec_end->next = sec;	/* then add new section at end of chain */
	}
	else
	{
		ini->sec = sec;	/* if chain does not exist, the new section is the beginning of the chain */
	}
}

void ini_rem_section(ini_file_t *ini, const char *secs)
{
	ini_section_t *sec, *old_sec = NULL;
	
	sec = ini->sec;
	while(sec)
	{
		if(strcmp_ncase(secs, sec->name) == 0)	/* if it's the section we're searching */
		{
			if(old_sec) { old_sec->next = sec->next; }	/* if it was a previous section, link it to the section next to the one to delete */
			else { ini->sec = sec->next; }	/* if it was no previous section, the section to delete is the beginning of chain */
			ini_free_section(sec);	/* delete it ! */
			return;
		}
		old_sec = sec;
		sec = sec->next;
	}
}

void ini_mod_section(ini_file_t *ini, const char *secs, const char *secs_new)
{
	ini_section_t *sec;
	
	sec = ini->sec;
	while(sec)
	{
		if(strcmp_ncase(secs, sec->name) == 0)	/* if it's the section we're searching */
		{
			strncpy(sec->name, secs_new, 64);	/* overwrite its name with new one */
			return;
		}
		sec = sec->next;
	}
}

ini_value_t *ini_end_value(ini_value_t *val)	/* returns end of a value chain */
{
	ini_value_t *val_stack = val, *val_stack_old = NULL;
	while(val_stack)
	{
		val_stack_old = val_stack;
		val_stack = val_stack->next;
	}
	return val_stack_old;
}

void ini_add_value(ini_file_t *ini, const char *secs, const char *vals, const char *value)
{
	ini_section_t *sec;
	ini_value_t *val, *val_end = NULL;
	
	sec = ini_req_section(ini, secs);	/* get section */
	if(!sec) return;
	
	val = (ini_value_t*)malloc(sizeof(ini_value_t));	/* allocate new value */
	if(!val) return;
	
	strncpy(val->name, vals, 64);	/* copy name */
	strncpy(val->value, value, 256);	/* copy value */
	val->next = NULL;	/* new value is at end of chain */
	
	if(sec->val)	/*check  if chain exists */
	{
		val_end = ini_end_value(sec->val);	/* get end of chain */
		val_end->next = val;	/* then add new section at end of chain */
	}
	else
	{
		sec->val = val;	/* if chain does not exist, the new section is the beginning of the chain */
	}
}

void ini_rem_value(ini_file_t *ini, const char *secs, const char *vals)
{
	ini_section_t *sec;
	ini_value_t *val, *old_val = NULL;
	
	sec = ini_req_section(ini, secs);	/* get section */
	if(!sec) return;
	
	val = sec->val;
	while(val)
	{
		if(strcmp_ncase(vals, val->name) == 0)	/* if it's the value we're searching */
		{
			if(old_val) { old_val->next = val->next; }	/* if it was a previous value, link it to the value next to the one to delete */
			else { sec->val = val->next; }	/* if it was no previous value, the value to delete is the beginning of chain */
			free(val);	/* delete it ! */
			return;
		}
		old_val = val;
		val = val->next;
	}
}

void ini_mod_value(ini_file_t *ini, const char *secs, const char *vals, const char *vals_new, const char *value_new)
{
	ini_section_t *sec;
	ini_value_t *val;
	
	sec = ini_req_section(ini, secs);	/* get section */
	if(!sec) return;
	
	val = sec->val;
	while(sec)
	{
		if(strcmp_ncase(vals, val->name) == 0)	/* if it's the section we're searching */
		{
			if(vals_new) strncpy(val->name, vals_new, 64);	/* overwrite its name with new one, if needed (vals_new != NULL) */
			if(value_new) strncpy(val->value, value_new, 256);	/* overwrite its value with new one, if needed (value_new != NULL) */
			return;
		}
		val = val->next;
	}
}

int ini_write(ini_file_t *ini, const char *filename)
{
	char line[512];
	FILE *file;
	ini_section_t *sec;
	ini_value_t *val;
	
	if(!ini) return -1;
	
	if(filename)
	{
		file = fopen(filename, "w+");
		if(!file) { return -1; }
	}
#ifndef __NDS__
	else { file = stdout; }
#else
	else { return -1; }
#endif
	
	sec = ini->sec;
	while(sec)
	{
		sprintf(line, "[%s]\r\n", sec->name);	/* writes section */
		fputs(line, file);
		
		val = sec->val;
		while(val)	/* for each value in section... */
		{
			sprintf(line, "%s = %s\r\n", val->name, val->value);	/* writes value */
			fputs(line, file);
			
			val = val->next;
		}
		
		fputs("\r\n", file);	/* add new line at end of section (looks better) */
		
		sec = sec->next;
	}
	
	if(filename) { fclose(file); }
	
	return 0;
}

void ini_log(ini_file_t *ini)
{
	ini_section_t *sec;
	ini_value_t *val;
	sec = ini->sec;
	while(sec)
	{
		fprintf(stderr, "[%s]\n", sec->name);
		val = sec->val;
		while(val)
		{
			fprintf(stderr, " + %s = %s\n", val->name, val->value);
			val = val->next;
		}
		sec = sec->next;
	}
}

#ifdef INI_TEST_MAIN
int main(int argc, char *argv[])
{
	ini_file_t *ini;
	
	ini = ini_parse(argv[1]);
	
	ini_section_t *sec;
	ini_value_t *val;
	sec = ini->sec;
	while(sec)
	{
		fprintf(stderr, "[%s]\n", sec->name);
		val = sec->val;
		while(val)
		{
			fprintf(stderr, " + %s = %s\n", val->name, val->value);
			val = val->next;
		}
		sec = sec->next;
	}
	
	const char *v;
	
	v = ini_req_value(ini, "sEcTiOn1", "vAlUe10");
	fprintf(stderr, "Req: %s\n", v);
	
	ini_add_section(ini, "added");
	ini_rem_section(ini, "section1");
	ini_mod_section(ini, "section2", "section1");
	
	ini_add_value(ini, "ADDED", "added_value", "SUPER!!!!!!");
	ini_add_value(ini, "ADDED", "added_value(deux)", "SUPER!!!!!! (biiiiiiiiiiiiiiissssssss)");
	
	ini_add_value(ini, "section1", "value11", "c'est la fete !");
	ini_rem_value(ini, "section1", "proutch");
	ini_mod_value(ini, "section1", "value20", "value10", "!!!!!!!");
	
	ini_write(ini, NULL);
	
	ini_free_file(ini);
	
	return 0;
}
#endif


